"""
OWT (One Way Trip)
Copyright (C) 2012 Juan J. Martinez <jjm@usebox.net>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
import math
import random

import pygame
from pygame.locals import *

from owt.data import load_png, load_sound

class Effect(pygame.sprite.Sprite):

    def __init__(self, x, y):
        super(Effect, self).__init__()

        self.sprites = load_png(self.resource_png)
        self.frames = []
        for f in range(self.nframes):
            self.frames.append(self.sprites.subsurface((f*64, 0, 64, 48)))

        self.frame = 0
        self.frame_delay = 80
        self.last_clock = 0

        self.x = x
        self.y = y

        fx = load_sound(self.resource_wav)
        fx.play()

        logging.debug("%s added: %d, %d" % (self.__class__, self.x, self.y))

    def update(self, clock):
        self.image = self.frames[self.frame]
        self.rect = self.image.get_rect()
        self.rect.move_ip(self.x, self.y)

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1
            if self.frame == len(self.frames):
                self.kill()
                return

            self.last_clock = clock


class ExplosionSmall(Effect):
    name = "explosion"
    resource_png = "explosion_small.png"
    resource_wav = "explosion_small.wav"
    nframes = 4

class Explosion(Effect):
    name = "explosion"
    resource_png = "explosion.png"
    resource_wav = "explosion.wav"
    nframes = 4

    def __init__(self, x, y):
        super(Explosion, self).__init__(x, y)

        self.frame_delay = 120

class PowerUp(Effect):
    name = "power-up"
    resource_png = "power_up.png"
    resource_wav = "up.wav"
    nframes = 3

    def __init__(self, x, y):
        super(PowerUp, self).__init__(x, y)

        self.frame_delay = 200

class Fire(pygame.sprite.Sprite):
    name = "fire"

    def __init__(self, x, y, efx, incx=6, incy=0, mute=False):
        super(Fire, self).__init__()

        self.efx = efx

        self.x = x
        self.y = y

        self.image = load_png("player_fire.png")
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y)

        self.incx = incx
        self.incy = incy

        if not mute:
            fx = load_sound('player_fire.wav')
            fx.play()

        logging.debug("%s added: %d, %d" % (self.__class__, self.x, self.y))

    def update(self, clock):
        self.x += self.incx
        self.y += self.incy

        if self.x < -20 or self.x > 650 or self.y < -20 or self.y > 410:
            self.kill()
            logging.debug("%s is done" % self.__class__)
            return

        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(round(self.x), round(self.y))

    def impact(self):
        logging.debug("%s impacted" % self.__class__)

        self.efx.add(ExplosionSmall(round(self.x)-4, round(self.y)-16))
        self.kill()

class BasePowerUp(pygame.sprite.Sprite):

    def __init__(self, x, y, player, efx):
        super(BasePowerUp, self).__init__()

        self.player = player
        self.efx = efx

        self.sprites = load_png(self.resource_png)
        self.frames = []
        for f in range(2):
            self.frames.append(self.sprites.subsurface((f*64, 0, 64, 48)))

        self.frame = 0
        self.frame_delay = 100
        self.last_clock = 0

        self.x = x
        self.y = y

        self.incx = -1

        self.image = self.frames[self.frame]
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y)

        logging.debug("%s added: %d, %d" % (self.__class__, self.x, self.y))

    def update(self, clock, px, py):
        self.x += self.incx

        if self.x < -60:
            self.kill()
            logging.debug("%s is done" % self.__class__)
            return

        self.image = self.frames[self.frame]
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y)

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1
            if self.frame == len(self.frames):
                self.frame = 0

            self.last_clock = clock

    def impact(self):
        logging.debug("%s impacted" % self.__class__)

        self.efx.add(PowerUp(self.x, self.y))
        self.kill()

class Shield(BasePowerUp):
    name = "shield"
    resource_png = "shield.png"

    def impact(self):
        super(Shield, self).impact()
        self.player.shield_up()

class Power(BasePowerUp):
    name = "power"
    resource_png = "power.png"

    def impact(self):
        super(Power, self).impact()
        self.player.power_up()

class EnemyFire(pygame.sprite.Sprite):
    name = "fire"

    def __init__(self, x, y, efx, incx=-4, incy=0):
        super(EnemyFire, self).__init__()

        self.efx = efx

        self.x = x
        self.y = y

        self.image = load_png("enemy_fire.png")
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y)

        self.incx = incx
        self.incy = incy

        fx = load_sound('enemy_fire.wav')
        fx.play()

        logging.debug("%s added: %d, %d" % (self.__class__, self.x, self.y))

    def update(self, clock, px, py):
        self.x += self.incx
        self.y += self.incy

        if self.x < -20 or self.x > 650 or self.y < -20 or self.y > 410:
            logging.debug("%s is done" % self.__class__)
            self.kill()
            return

        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(math.floor(self.x), math.floor(self.y))

    def impact(self):
        logging.debug("%s impacted" % self.__class__)

        self.efx.add(ExplosionSmall(round(self.x)-28, round(self.y)-20))
        self.kill()

class Player(pygame.sprite.Sprite):
    name = "player"

    def __init__(self, x, y, efx):
        super(Player, self).__init__()

        self.efx = efx
        self.sprites = load_png("player.png")
        self.frames = []
        for f in range(2):
            self.frames.append(self.sprites.subsurface((f*64, 0, 64, 48)))

        self.frame = 0
        self.frame_delay = 60
        self.last_clock = 0

        self.x = x
        self.y = y

        self.incx = 0
        self.incy = 0

        self.top_hit = 5
        self.hit = self.top_hit

        self.power = 0
        self.fire_power = 4 # 1 is top speed
        self.fire_delay = self.fire_power
        self.fire_alt = 0
        self.fire = False

        logging.debug("Player added: %d, %d" % (self.x, self.y))

    def up(self):
        self.incy = -3
        self.stopy = False

    def down(self):
        self.incy = 3
        self.stopy = False

    def left(self):
        self.incx = -3
        self.stopx = False

    def right(self):
        self.incx = 3
        self.stopx = False

    def stop_vert(self):
        self.stopy = True

    def stop_horiz(self):
        self.stopx = True

    def impact(self):
        self.hit -= 1
        if self.hit <= 0 and self.alive():
            logging.debug("%s was destroyed" % self.__class__)
            self.efx.add(Explosion(self.x, self.y))
            self.kill()

            self.x = -100
            self.y = -100

    def power_up(self):
        self.power += 1

        if self.power & 1:
            if self.fire_power > 1:
                self.fire_power -= 1

    def shield_up(self):
        if self.hit < self.top_hit:
            self.hit += 1

    def update(self, clock):
        self.x += self.incx
        self.y += self.incy

        if self.x < 20:
            self.x = 20
        elif self.x > 580:
            self.x= 580
        if self.y < 20:
            self.y = 20
        elif self.y > 332:
            self.y = 332

        self.image = self.frames[self.frame]
        self.rect = self.image.get_bounding_rect()
        if self.frame == 0:
            # fix for the animation
            self.rect[0] += 1
        self.rect.move_ip(self.x, self.y)

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1
            if self.frame == len(self.frames):
                self.frame = 0

            if self.stopx and self.incx != 0:
                self.incx = self.incx + 1 if self.incx < 0 else self.incx -1
            if self.stopy and self.incy != 0:
                self.incy = self.incy + 1 if self.incy < 0 else self.incy -1

            if self.fire:
                self.fire_delay -= 1
                if self.fire_delay == 0:
                    self.fire_delay = self.fire_power
                    self.fire_alt = 1 if not self.fire_alt else 0
                    if self.power < 2:
                        fire = Fire(self.x+34, self.y+32, self.efx)
                        fire.update(clock)
                        self.groups()[0].add(fire)
                    elif self.power == 2:
                        fire = Fire(self.x+34, self.y+32, self.efx, incy = 0.2)
                        fire.update(clock)
                        self.groups()[0].add(fire)
                        fire = Fire(self.x+34, self.y+32, self.efx, incy = -0.2, mute = True)
                        fire.update(clock)
                        self.groups()[0].add(fire)
                    elif self.power == 3:
                        fire = Fire(self.x+34, self.y+32, self.efx)
                        fire.update(clock)
                        self.groups()[0].add(fire)
                        if self.fire_alt == 0:
                            fire = Fire(self.x+34, self.y+32, self.efx, incy = 0.2, mute = True)
                            fire.update(clock)
                            self.groups()[0].add(fire)
                            fire = Fire(self.x+34, self.y+32, self.efx, incy = -0.2, mute = True)
                            fire.update(clock)
                            self.groups()[0].add(fire)
                    elif self.power >= 4:
                        fire = Fire(self.x+34, self.y+32, self.efx)
                        fire.update(clock)
                        self.groups()[0].add(fire)
                        if self.fire_alt == 0:
                            fire = Fire(self.x+34, self.y+32, self.efx, incy = 0.2, mute = True)
                            fire.update(clock)
                            self.groups()[0].add(fire)
                            fire = Fire(self.x+34, self.y+32, self.efx, incy = -0.2, mute = True)
                            fire.update(clock)
                            self.groups()[0].add(fire)
                        else:
                            fire = Fire(self.x+34, self.y+32, self.efx, incy = 0.3, mute = True)
                            fire.update(clock)
                            self.groups()[0].add(fire)
                            fire = Fire(self.x+34, self.y+32, self.efx, incy = -0.3, mute = True)
                            fire.update(clock)
                            self.groups()[0].add(fire)


            self.last_clock = clock

class BaseEnemy(pygame.sprite.Sprite):
    name = "enemy"

    # does this enemy provide power ups?
    power = False
    shield= False

    def __init__(self, x, y, efx):
        super(BaseEnemy, self).__init__()

        self.efx = efx
        self.sprites = load_png(self.resource_png)
        self.frames = []
        for f in range(self.nframes):
            self.frames.append(self.sprites.subsurface((f*64, 0, 64, 48)))

        self.frame = 0
        self.last_clock = 0

        self.x = x
        self.y = y

        self.fire_delay = 0

        logging.debug("%s added: %d, %d" % (self.__class__, self.x, self.y))

    def impact(self):
        self.hits -= 1
        if self.hits <= 0 and self.alive():
            logging.debug("%s was destroyed" % self.__class__)
            self.efx.add(Explosion(self.rect[0], self.rect[1]))
            self.kill()

class EnemyShip(BaseEnemy):
    score = 15
    resource_png = "enemy_ship.png"
    nframes = 4

    def __init__(self, x, y, efx):
        super(EnemyShip, self).__init__(x, y, efx)

        self.frame_delay = 80

        self.incx = -2
        self.incy = 0

        self.hits = 2
        self.fire = 80

        self.back = 0

    def update(self, clock, px, py):
        self.x += self.incx

        if self.fire_delay < self.fire:
            self.fire_delay += 1
        elif not self.back and px+32 < self.x and py+24 > self.y+22 and py+24 < self.y+28:
            self.fire_delay = 0
            self.groups()[0].add(EnemyFire(self.rect[0], self.rect[1]+4+self.rect[3]/2, self.efx, incx = self.incx*2))
            logging.debug("%s fired!" % self.__class__)

        self.image = self.frames[self.frame + self.back]
        self.rect = self.image.get_bounding_rect()
        if self.back:
            if self.frame == 1:
                # fix for the animation
                self.rect[0] += 2
        self.rect.move_ip(self.x, self.y+(24-self.rect[1]))

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1
            if self.frame == 2:
                self.frame = 0

            self.last_clock = clock

        if self.back and self.x > 660:
            self.kill()
            logging.debug("%s is done" % self.__class__)
            return

        if self.x < -100 and not self.back:
            self.back = 2
            self.incx = 3

class EnemyDisk(BaseEnemy):
    score = 10
    resource_png = "enemy_disk.png"
    nframes = 8

    def __init__(self, x, y, efx):
        super(EnemyDisk, self).__init__(x, y, efx)

        self.frame_delay = 160

        self.incx = -1
        self.incy = 0

        self.hits = 2
        self.fire = 80

    def update(self, clock, px, py):
        self.x += self.incx
        self.y += self.incy

        if self.x < 500 and self.incy == 0:
            if self.y < 200:
                self.incy = 2
            else:
                self.incy = -2

        if self.fire_delay < self.fire:
            self.fire_delay += 1
        elif px+32 < self.x and py+24 > self.y+22 and py+24 < self.y+28:
            self.fire_delay = 0
            self.groups()[0].add(EnemyFire(self.x+14, self.rect[1]+4+self.rect[3]/2, self.efx))
            logging.debug("%s fired!" % self.__class__)

        if self.y < 10 or self.y > 320:
            self.incy *= -1

        self.image = self.frames[self.frame]
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y+(24-self.rect[1]))

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1
            if self.frame == len(self.frames):
                self.frame = 0

            self.last_clock = clock

        if self.x < -60 or self.x > 660 or self.y > 420 or self.y < -20:
            self.kill()
            logging.debug("%s is done" % self.__class__)
            return

class EnemyDisk2(EnemyDisk):
    score = 5

    def __init__(self, x, y, efx):
        super(EnemyDisk2, self).__init__(x, y, efx)

        self.fire = 80

    def update(self, clock, px, py):
        self.x += self.incx

        if self.fire_delay < self.fire:
            self.fire_delay += 1
        elif px+32 < self.x and py+24 > self.y+22 and py+24 < self.y+28:
            self.fire_delay = 0
            self.groups()[0].add(EnemyFire(self.x+14, self.rect[1]+4+self.rect[3]/2, self.efx))
            logging.debug("%s fired!" % self.__class__)

        self.image = self.frames[self.frame]
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y+(24-self.rect[1]))

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1
            if self.frame == len(self.frames):
                self.frame = 0

            self.last_clock = clock

        if self.x < -60 or self.x > 660 or self.y > 420 or self.y < -20:
            self.kill()
            logging.debug("%s is done" % self.__class__)
            return

class EnemyEye(BaseEnemy):
    score = 10
    resource_png = "enemy_eye.png"
    nframes = 5

    def __init__(self, x, y, efx):
        super(EnemyEye, self).__init__(x, y, efx)

        self.frame_delay = 180

        self.opening = True
        self.circle = None
        self.r = 150
        self.rad = -180.0
        self.fire = True

        self.incx = -3
        self.incy = 0

        self.hits = 4

    def update(self, clock, px, py):

        if self.circle:
            self.x = math.floor(self.r*math.cos(self.rad/100)+300)
            self.y = math.floor(self.r*math.sin(self.rad/100)+160)
            self.rad -= 2
            if self.rad < -1438:
                self.circle = False
        else:
            self.x += self.incx

        if self.x < 264 and self.circle is None:
            self.circle = True

        self.image = self.frames[self.frame]
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y+(24-self.rect[1]))

        if self.frame == 4 and not self.fire and all([px > 0, py > 0]):
            if py != self.rect[1]:
                dx = (px-self.rect[0])/float(abs(py-self.rect[1]))
            else:
                dx = 0
            if px != self.rect[0]:
                dy = (py-self.rect[1])/float(abs(px-self.rect[0]))
            else:
                dy = 0
            if abs(dy) < 1:
                dy *= 3
                if self.x < px:
                    dx = 3
                else:
                    dx = -3
                self.fire = True
            elif abs(dx) < 1:
                dx *= 3
                if self.y < py:
                    dy = 3
                else:
                    dy = -3
                self.fire = True

            if self.fire:
                self.groups()[0].add(EnemyFire(self.rect[0]+22, self.rect[1]+22, self.efx, incx = dx, incy = dy))
                logging.debug("%s fired!" % self.__class__)

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1 if self.opening else -1
            if self.frame == len(self.frames):
                self.opening = False
                self.frame -= 1
            elif self.frame < 0:
                self.fire = False
                self.opening = True
                self.frame += 1

            self.last_clock = clock

        if self.x < -60 or self.x > 660 or self.y > 420 or self.y < -20:
            self.kill()
            logging.debug("%s is done" % self.__class__)
            return

class EnemyEyeInner(EnemyEye):
    score = 10
    resource_png = "enemy_eye.png"
    nframes = 5

    def update(self, clock, px, py):
        super(EnemyEyeInner, self).update(clock, px, py)

        if self.circle and self.r >= 90:
            self.r -= 2

# FIXME: DRY!
class EnemyEye2(EnemyEye):
    score = 10
    resource_png = "enemy_eye.png"
    nframes = 5

    def __init__(self, x, y, efx):
        super(EnemyEye2, self).__init__(x, y, efx)

        self.rad = 176.0

    def update(self, clock, px, py):

        if self.circle:
            self.x = math.floor(self.r*math.cos(self.rad/100)+300)
            self.y = math.floor(self.r*math.sin(self.rad/100)+160)
            self.rad += 2
            if self.rad > 1438:
                self.circle = False
        else:
            self.x += self.incx

        if self.x < 274 and self.circle is None:
            self.circle = True

        self.image = self.frames[self.frame]
        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y+(24-self.rect[1]))

        if self.frame == 4 and not self.fire and all([px > 0, py > 0]):
            if py != self.rect[1]:
                dx = (px-self.rect[0])/float(abs(py-self.rect[1]))
            else:
                dx = 0
            if px != self.rect[0]:
                dy = (py-self.rect[1])/float(abs(px-self.rect[0]))
            else:
                dy = 0
            if abs(dy) < 1:
                dy *= 3
                if self.x < px:
                    dx = 3
                else:
                    dx = -3
                self.fire = True
            elif abs(dx) < 1:
                dx *= 3
                if self.y < py:
                    dy = 3
                else:
                    dy = -3
                self.fire = True

            if self.fire:
                self.groups()[0].add(EnemyFire(self.rect[0]+22, self.rect[1]+22, self.efx, incx = dx, incy = dy))
                logging.debug("%s fired!" % self.__class__)

        if clock - self.last_clock > self.frame_delay:
            self.frame += 1 if self.opening else -1
            if self.frame == len(self.frames):
                self.opening = False
                self.frame -= 1
            elif self.frame < 0:
                self.fire = False
                self.opening = True
                self.frame += 1

            self.last_clock = clock

        if self.x < -60 or self.x > 660 or self.y > 420 or self.y < -20:
            self.kill()
            logging.debug("%s is done" % self.__class__)
            return

class Meteor(pygame.sprite.Sprite):
    name = "enemy"

    power = False
    shield= False
    score = 0

    def __init__(self, x, y, efx):
        super(Meteor, self).__init__()

        self.image = load_png("meteor.png")
        self.image = pygame.transform.rotate(self.image, random.randint(0, 25))

        self.x = y
        self.y = -128

        self.incx = -2
        self.incy = 2

        logging.debug("%s added: %d, %d" % (self.__class__, self.x, self.y))

    def update(self, clock, px, py):
        self.x += self.incx
        self.y += self.incy

        self.rect = self.image.get_bounding_rect()
        self.rect.move_ip(self.x, self.y)

        if self.y > 440:
            logging.debug("%s is done" % self.__class__)
            self.kill()
            return

    def impact(self):
        # can't be destroyed!
        # sorry about that
        pass

