"""
OWT (One Way Trip)
Copyright (C) 2012 Juan J. Martinez <jjm@usebox.net>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import json
import logging

import pygame
from pygame.locals import *

_data_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')
_cache = {}

def package_path(fn):
    def translate_filename(filename, *args):
        filename = os.path.join(_data_path, filename)
        return fn(filename, *args)
    return translate_filename

@package_path
def load_json(filename):
    obj = get(filename)
    if obj:
        return obj

    with open(filename, 'r') as fd:
        obj = json.load(fd)

    put(filename, obj)
    return obj

@package_path
def load_png(filename):
    image = get(filename)
    if image:
        return image

    image = pygame.image.load(filename)
    if image.get_alpha is None:
        image = image.convert()
    else:
        image = image.convert_alpha()

    put(filename, image)
    return image

@package_path
def load_sound(filename):
    sound = get(filename)
    if sound:
        return sound

    sound = pygame.mixer.Sound(filename)

    put(filename, sound)
    return sound

@package_path
def load_music(filename):
    # it doesn't use the cache, but it's coherent to put it here
    pygame.mixer.music.load(filename)
    return None

@package_path
def load_font(filename, size):
    key = "%s:%s" % (filename, size)
    font = get(key)
    if font:
        return font

    font = pygame.font.Font(filename, size)
    font.set_bold(True)

    put(key, font)
    return font

@package_path
def free_resource(filename):
    delete(filename)

def get(key):
    if key in _cache.keys():
        logging.debug("Cache hit: %s" % key)
        return _cache[key]
    else:
        logging.debug("Cache miss: %s" % key)
        return None

def put(key, value):
    logging.debug("Cache put: %s" % key)
    _cache[key] = value

def delete(key):
    if key in _cache.keys():
        logging.debug("Cache delete: %s" % key)
        del _cache[key]

