"""
OWT (One Way Trip)
Copyright (C) 2012 Juan J. Martinez <jjm@usebox.net>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import optparse
import logging

import pygame
from pygame.locals import *

from owt.constants import *
from owt.data import load_png, load_sound, load_font, free_resource, load_json, load_music
from owt.actors import Player
import owt.actors

class Main(object):

    def __init__(self):
        parser = optparse.OptionParser(description=project_descr,
                                       epilog='Project home: ' + project_url,
                                       version='%prog ' + version)
        parser.add_option("-f", "--fullscreen", dest="fullscreen", default=False,
                          action="store_true")
        parser.add_option("-d", "--debug", dest="debug", default=False,
                          action="store_true")
        self.options, args = parser.parse_args()

        if self.options.debug:
            logging.basicConfig(level=logging.DEBUG)
            logging.debug("Debug enabled")
            logging.debug("Options: %s" % self.options)

    def run(self):
        pygame.mixer.pre_init(frequency=44100)
        pygame.init()

        flags = 0
        if self.options.fullscreen:
            flags = pygame.DOUBLEBUF | pygame.FULLSCREEN | pygame.HWSURFACE

        self.screen = pygame.display.set_mode((640, 400), flags)
        pygame.display.set_caption(app_name)
        pygame.mouse.set_visible(False)

        self.sw, self.sh = self.screen.get_size()

        # TODO: save on disk!
        hiscore = 0

        if not self.options.debug:
            self.screen_usebox()

        while True:
            option = self.screen_main_menu(hiscore)
            if option == 'quit':
                self.screen.fill((0, 0, 0))
                pygame.display.flip()
                break

            score = self.screen_game()
            if score > hiscore:
                hiscore = score

    def screen_usebox(self):
        """Display usebox logo and play a sound"""

        logo = load_png('useboxnet.png')
        sound = load_sound('useboxnet.wav')

        bg = pygame.Surface(self.screen.get_size())
        bg = bg.convert()
        bg.fill((0, 0, 0))

        self.screen.blit(bg, (0, 0))

        w, h = logo.get_size()
        self.screen.blit(logo, (self.sw/2-w/2, self.sh/2-h/2))
        pygame.display.flip()
        sound.play()

        pygame.time.delay(3000)

        # we're not using these anymore
        free_resource('useboxnet.png')
        free_resource('useboxnet.wav')

    def screen_main_menu(self, hiscore):
        """Display the main menu"""

        bg = load_png('main_menu.png')
        self.screen.blit(bg, (0, 0))

        font = load_font(FONT, 18)
        t = font.render("Press s to start", True, (255, 255, 255))
        self.screen.blit(t, (self.sw/2-t.get_width()/2, 260))

        font = load_font(FONT, 11)
        t = font.render("HI-SCORE: %.5d" % hiscore, True, (200, 200, 200))
        self.screen.blit(t, (self.sw/2-t.get_width()/2, 380))

        pygame.display.flip()

        while True:
            event = pygame.event.wait()
            if event.type == QUIT:
                return 'quit'
            elif event.type == KEYDOWN:
                if event.key in [K_ESCAPE, K_q]:
                    return 'quit'
                elif event.key == K_s:
                    return 'start'

    def screen_game(self):
        """The game loop(tm)"""

        script = load_json("script.json")
        script_pointer = 0
        script_delay = 0

        try:
            load_music("owt.ogg")
        except Exception, e:
            logging.debug(e)
            music = False
        else:
            music = True
            pygame.mixer.music.play(-1)

        font = load_font(FONT, 11)
        display_font = load_font(FONT, 30)

        display_text = ""
        display_text_delay = 0

        score = 0
        dead_delay = 800

        hits = load_png("hits.png")
        game_over = load_png("game_over.png")

        sf = load_png("starfield.png")
        sf_scroll = 0
        sf_scroll_t = 0
        self.screen.blit(sf, (0, 0, 640, 400))

        clock = pygame.time.get_ticks()

        efx_group = pygame.sprite.LayeredUpdates()

        player_group = pygame.sprite.LayeredUpdates()
        player = Player(100, 220, efx_group)
        player_group.add(player)

        enemy_group = pygame.sprite.LayeredUpdates()

        while True:
            if pygame.event.peek(QUIT):
                pygame.event.clear()
                return 0

            keys = pygame.key.get_pressed()

            if keys[K_ESCAPE] or keys[K_q]:
                pygame.mixer.music.stop()
                pygame.event.clear()
                return 0

            if keys[K_UP]:
                player.up()
            if keys[K_DOWN]:
                player.down()
            if keys[K_RIGHT]:
                player.right()
            if keys[K_LEFT]:
                player.left()
            if keys[K_SPACE]:
                player.fire = True
            else:
                player.fire = False

            if not any([keys[K_UP], keys[K_DOWN]]) or all([keys[K_UP], keys[K_DOWN]]):
                player.stop_vert()
            if not any([keys[K_LEFT], keys[K_RIGHT]]) or all([keys[K_LEFT], keys[K_RIGHT]]):
                player.stop_horiz()

            new_clock = pygame.time.get_ticks()

            if new_clock-clock > 1000/FPS:

                # the starfield
                if new_clock-sf_scroll_t > 1000/(FPS/3):
                    sf_scroll_t = new_clock
                    if sf_scroll > 1280:
                        sf_scroll = 0
                    else:
                        sf_scroll += 1

                    sf.blit(sf, (1278, 0, 2, 400), (0, 0, 2, 400))
                    sf.scroll(-1)

                self.screen.blit(sf, (0, 0, 640, 400))

                enemy_group.update(new_clock, player.x, player.y)
                enemy_group.draw(self.screen)

                player_group.update(new_clock)
                player_group.draw(self.screen)

                efx_group.update(new_clock)
                efx_group.draw(self.screen)

                # FIXME: oh, please... more planning next time
                for enemy in [enemy for enemy in enemy_group if enemy.name != 'fire']:
                    impacts = pygame.sprite.spritecollide(enemy, player_group, False, pygame.sprite.collide_mask)
                    for item in impacts:
                        if item != player and enemy.name not in [ 'power', 'shield' ]:
                            score += enemy.score
                            item.impact()
                            enemy.impact()
                            if not enemy.alive() and enemy.power:
                                enemy_group.add(owt.actors.Power(enemy.rect[0], enemy.rect[1], player, efx_group))
                            elif not enemy.alive() and enemy.shield:
                                enemy_group.add(owt.actors.Shield(enemy.rect[0], enemy.rect[1], player, efx_group))

                if player.alive():
                    impacts = pygame.sprite.spritecollide(player, enemy_group, False, pygame.sprite.collide_mask)
                    for enemy in impacts:
                        enemy.impact()
                        # FIXME: use a differen group for the power-ups!
                        if enemy.name not in ["power", "shield"]:
                            player.impact()

                t = font.render("%.5d" % score, True, (255, 255, 255))
                self.screen.blit(t, (8, 385))

                self.screen.blit(hits,(14+t.get_width(), 388), (0, 0, 300, 8))
                if player.alive():
                    self.screen.blit(hits,(14+t.get_width(), 388), (0, 8, 300*player.hit/player.top_hit, 8))
                else:
                    self.screen.blit(game_over,(self.sw/2-game_over.get_width()/2, self.sh/2-game_over.get_height()))
                    dead_delay -= 1
                    if music:
                        pygame.mixer.music.fadeout(3000)
                        music = False
                    if dead_delay < 0:
                        return score

                if display_text_delay:
                    display_text_delay -= 1
                    t = display_font.render(display_text, True, (255, 142, 43))
                    self.screen.blit(t, (self.sw/2-t.get_width()/2, 0))
                    t = display_font.render(display_text, True, (183, 31, 0))
                    self.screen.blit(t, (self.sw/2-t.get_width()/2, 2))

                clock = new_clock
                pygame.display.flip()

                # play the script!
                if script_delay == script[script_pointer]['delay']:
                    script_delay = 0
                    if script[script_pointer]['event'] == 'END':
                        # oh, well... One Way Trip
                        player.hit = 1
                        player.impact()
                    elif script[script_pointer]['event'] == 'DISPLAY':
                        display_text = script[script_pointer]['text']
                        display_text_delay = script[script_pointer]['text_delay']
                        script_pointer += 1
                    else:
                        # This is POWAH! Love Python
                        cls = getattr(owt.actors, script[script_pointer]['event'])
                        grp = locals()["%s_group" % script[script_pointer]['type']]
                        y = script[script_pointer]['position']
                        item = cls(650, y, efx_group)
                        if script[script_pointer].get("power", False):
                            item.power = True
                        if script[script_pointer].get("shield", False):
                            item.shield = True
                        grp.add(item)
                        script_pointer += 1
                else:
                    if player.alive():
                        script_delay += 1

            else:
                pygame.time.wait((1000/FPS)-(pygame.time.get_ticks()-clock))

